# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/add_repo.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(364, 102)
        Dialog.setModal(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.source = QtWidgets.QLineEdit(Dialog)
        self.source.setObjectName("source")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.source)
        self.title = QtWidgets.QLineEdit(Dialog)
        self.title.setObjectName("title")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.title)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.verticalLayout.addLayout(self.formLayout)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.okButton = QtWidgets.QPushButton(Dialog)
        self.okButton.setDefault(True)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        self.cancelButton = QtWidgets.QPushButton(Dialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Add a new mod source"))
        self.label_2.setText(_translate("Dialog", "Source:"))
        self.label_3.setText(_translate("Dialog", "Title:"))
        self.okButton.setText(_translate("Dialog", "OK"))
        self.cancelButton.setText(_translate("Dialog", "Cancel"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_help.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.label.setWordWrap(True)
        self.label.setOpenExternalLinks(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "<html><head/><body><p>If you need help, you can either check <a href=\"http://www.hard-light.net/forums/index.php?topic=86364.msg1768735#msg1768735\"><span style=\" text-decoration: underline; color:#0000ff;\">this release post</span></a> or <a href=\"https://github.com/ngld/knossos/issues\"><span style=\" text-decoration: underline; color:#0000ff;\">check the reported issues</span></a>.</p></body></html>"))




## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function, division

import sys
import os
import logging
import subprocess
import struct
import hashlib
import tempfile
import re
import time
import json
import random
import functools
import glob
import semantic_version
import requests
from collections import OrderedDict
from threading import Condition, Event
from collections import deque

from . import center, progress

try:
    from PIL import Image
except ImportError:
    Image = None

SEVEN_PATH = '7z'
# Copied from http://sourceforge.net/p/sevenzipjbind/code/ci/master/tree/jbinding-java/src/net/sf/sevenzipjbinding/ArchiveFormat.java
# to conform to the FSO Installer.
ARCHIVE_FORMATS = ('zip', 'tar', 'split', 'rar', 'lzma', 'iso', 'hfs', 'gzip', 'gz',
                   'cpio', 'bzip2', 'bz2', '7z', 'z', 'arj', 'cab', 'lzh', 'chm',
                   'nsis', 'deb', 'rpm', 'udf', 'wim', 'xar')
# TODO: Too many?
USER_AGENTS = (
    'Mozilla/5.0 (compatible; Knossos %s +http://dev.tproxy.de/knossos/)' % (center.VERSION),
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/5.0)',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; Trident/6.0)',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; WOW64; Trident/6.0)',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0)',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.78.2 (KHTML, like Gecko) Version/6.1.6 Safari/537.78.2',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.124 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; de-de) AppleWebKit/125.2 (KHTML, like Gecko)',
    'Mozilla/5.0 (Windows NT 5.0; rv:21.0) Gecko/20100101 Firefox/21.0',
    'Mozilla/5.0 (Windows NT 5.1; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Windows NT 5.1; rv:6.0.2) Gecko/20100101 Firefox/6.0.2',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/38.0.2125.101 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; rv:32.0.3) Gecko/20100101 Firefox/32.0.3 anonymized by Abelssoft 1108446737',
    'Mozilla/5.0 (Windows NT 6.1; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/38.0.2125.104 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.4 (KHTML, like Gecko) Chrome/22.0.1229.94 Safari/537.4',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0',
    'Mozilla/5.0 (Windows NT 6.2; WOW64; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:29.0) Gecko/20100101 Firefox/29.0 SeaMonkey/2.26.1',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:31.0) Gecko/20100101 Firefox/31.0 DT-Browser/DTB7.031.0020',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:32.0) Gecko/20100101 Firefox/32.0',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; LCJB; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows; U; MSIE 9.0; Windows NT 9.0; en-US)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; de-AT; rv:1.7.12)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; de; rv:1.9.0.10) Gecko/2009042316 Firefox/3.0.10 (.NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; de; rv:1.9.2.24) Gecko/20111103 Firefox/3.6.24 ( .NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; de; rv:1.9.2.8) Gecko/20100722 Firefox/3.6.8 (.NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:x.x.x) Gecko/20041107 Firefox/x.x',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; pl; rv:1.9.1.3) Gecko/20090824 Firefox/3.5.3',
    'Mozilla/5.0 (Windows; U; Windows NT 6.0; de; rv:1.9.2.16) Gecko/20110319 BTRS26718 Firefox/3.6.16 (.NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.9.0.1) Gecko/2008070208 Firefox/3.0.1',
    'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.4 (KHTML, like Gecko) Iron/5.0.381.0 Chrome/5.0.381 Safari/533.4',
    'Mozilla/5.0 (Windows; U; Windows NT 6.1; tr-TR) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27',
    'Mozilla/5.0 (X11; Linux i686) AppleWebKit/534.7 (KHTML, like Gecko) Version/5.0 Safari/534.7',
    'Mozilla/5.0 (X11; Linux i686; rv:21.0) Gecko/20100101 Firefox/21.0',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2185.0 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:31.0) Gecko/20100101 Firefox/31.0',
    'Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:14.0; ips-agent) Gecko/20100101 Firefox/14.0.1',
    'Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:15.0) Gecko/20100101 Firefox/15.0.1',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:21.0) Gecko/20100101 Firefox/21.0',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:22.0) Gecko/20100101 Firefox/22.0',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:33.0) Gecko/20100101 Firefox/33.0',
    'Opera/9.80 (Windows NT 6.1; U; en) Presto/2.8.131 Version/11.10',
    'Opera/9.80 (Windows NT 6.1; WOW64) Presto/2.12.388 Version/12.17'
)
HTTP_SESSION = requests.Session()
HTTP_SESSION.verify = True
QUIET = not center.DEBUG
QUIET_EXC = False
HASH_CACHE = dict()
_HAS_CONVERT = None
_HAS_TAR = None
DL_POOL = None
_DL_CANCEL = Event()
_DL_CANCEL.clear()


# See code/cmdline/cmdline.cpp (in the SCP source) for details on the data structure.
class FlagsReader(object):
    _stream = None
    easy_flags = None
    flags = None
    
    def __init__(self, stream):
        self._stream = stream
        self.read()
    
    def unpack(self, fmt):
        if isinstance(fmt, struct.Struct):
            return fmt.unpack(self._stream.read(fmt.size))
        else:
            return struct.unpack(fmt, self._stream.read(struct.calcsize(fmt)))
    
    def read(self):
        # Explanation of unpack() and Struct() parameters: http://docs.python.org/3/library/struct.html#format-characters
        self.easy_flags = OrderedDict()
        self.flags = OrderedDict()
        
        easy_size, flag_size = self.unpack('2i')
        
        easy_struct = struct.Struct('32s')
        flag_struct = struct.Struct('20s40s?ii16s256s')
        
        if easy_size != easy_struct.size:
            logging.error('EasyFlags size is %d but I expected %d!', easy_size, easy_struct.size)
            return
        
        if flag_size != flag_struct.size:
            logging.error('Flag size is %d but I expected %d!', flag_size, flag_struct.size)
            return
        
        for i in range(self.unpack('i')[0]):
            self.easy_flags[1 << i] = self.unpack(easy_struct)[0].decode('utf8').strip('\x00')
        
        for i in range(self.unpack('i')[0]):
            flag = self.unpack(flag_struct)
            flag = {
                'name': flag[0].decode('utf8').strip('\x00'),
                'desc': flag[1].decode('utf8').strip('\x00'),
                'fso_only': flag[2],
                'on_flags': flag[3],
                'off_flags': flag[4],
                'type': flag[5].decode('utf8').strip('\x00'),
                'web_url': flag[6].decode('utf8').strip('\x00')
            }
            
            if flag['type'] not in self.flags:
                self.flags[flag['type']] = []
            
            self.flags[flag['type']].append(flag)


class ResizableSemaphore(object):
    _capacity = 0
    _free = 0
    _cond = None

    def __init__(self, cap):
        self._capacity = cap
        self._free = cap
        self._cond = Condition()

    def acquire(self, blocking=True, timeout=None):
        if not blocking and timeout is not None:
            raise ValueError("can't specify timeout for non-blocking acquire")

        rc = False
        endtime = None
        with self._cond:
            while self._free < 1:
                if not blocking:
                    break
                if timeout is not None:
                    if endtime is None:
                        endtime = time.time() + timeout
                    else:
                        timeout = endtime - time.time()
                        if timeout <= 0:
                            break
                self._cond.wait(timeout)
            else:
                self._free -= 1
                rc = True
        return rc

    def release(self):
        with self._cond:
            if self._free >= self._capacity:
                raise ValueError('Semaphore released too many times')

            self._free += 1
            if self._free > 0:
                self._cond.notify()

    __enter__ = acquire

    def __exit__(self, t, v, tb):
        self.release()

    def get_capacity(self):
        return self._capacity

    def set_capacity(self, cap):
        diff = cap - self._capacity
        with self._cond:
            self._capacity = cap
            self._free += diff

            self._cond.notify_all()

        # logging.debug('Capacity set to %d, was %d. I now have %d free slots.', self._capacity, self._capacity - diff, self._free)
    
    def get_consumed(self):
        with self._cond:
            return self._capacity - self._free


class SpeedCalc(object):
    speeds = None
    last_time = 0
    last_bytes = 0
    interval = 0.5

    def __init__(self):
        self.speeds = deque(maxlen=30)

    def push(self, bytes):
        now = time.time()

        if self.last_time != 0:
            if (now - self.last_time) < self.interval:
                # Enforce a little gap between two speed measurements.
                return -1

            self.speeds.append(float(bytes - self.last_bytes) / float(now - self.last_time))

        self.last_time = now
        self.last_bytes = bytes

    def get_speed(self):
        if len(self.speeds) == 0:
            return 0.1

        return max(sum(self.speeds) / len(self.speeds), 0.1)


def call(*args, **kwargs):
    if sys.platform.startswith('win') and not center.DEBUG:
        # Provide the called program with proper I/O on Windows.
        if 'stdin' not in kwargs:
            kwargs['stdin'] = subprocess.DEVNULL
        
        if 'stdout' not in kwargs:
            kwargs['stdout'] = subprocess.DEVNULL
        
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.DEVNULL
        
        si = subprocess.STARTUPINFO()
        si.dwFlags = subprocess.STARTF_USESHOWWINDOW
        si.wShowWindow = subprocess.SW_HIDE
        
        kwargs['startupinfo'] = si

    logging.debug('Running %s', args[0])
    return subprocess.call(*args, **kwargs)


def check_output(*args, **kwargs):
    if sys.platform.startswith('win'):
        # Provide the called program with proper I/O on Windows.
        if 'stdin' not in kwargs:
            kwargs['stdin'] = subprocess.DEVNULL
        
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.DEVNULL
        
        si = subprocess.STARTUPINFO()
        si.dwFlags = subprocess.STARTF_USESHOWWINDOW
        si.wShowWindow = subprocess.SW_HIDE
        
        kwargs['startupinfo'] = si

    if 'universal_newlines' not in kwargs:
        kwargs['universal_newlines'] = True

    logging.debug('Running %s', args[0])
    return subprocess.check_output(*args, **kwargs)


def format_bytes(value):
    unit = 'Bytes'
    if value > 1024:
        value = value / 1024
        unit = 'KB'
    if value > 1024:
        value = value / 1024
        unit = 'MB'
    if value > 1024:
        value = value / 1024
        unit = 'GB'
    if value > 1024:
        value = value / 1024
        unit = 'TB'

    return str(round(value)) + ' ' + unit


def get(link, headers=None, random_ua=False, raw=False):
    global HTTP_SESSION

    if random_ua:
        if headers is None:
            headers = {}

        headers['User-Agent'] = get_user_agent(True)

    result = None
    try:
        result = HTTP_SESSION.get(link, headers=headers)
        if result.status_code == 304:
            return 304
        elif result.status_code != 200:
            result.raise_for_status()
    except:
        if result is None:
            logging.exception('Failed to load "%s"!', link)
        else:
            logging.error('Failed to load "%s"! (%d %s)', link, result.status_code, result.reason)
        return None

    if raw:
        return result
    else:
        return result.text


def post(link, data, headers=None, random_ua=False):
    global HTTP_SESSION

    if random_ua:
        if headers is None:
            headers = {}

        headers['User-Agent'] = get_user_agent(True)

    result = None
    try:
        result = HTTP_SESSION.post(link, data=data, headers=headers)
        if result.status_code != 200:
            result.raise_for_status()
    except:
        if result is None:
            logging.exception('Failed to load "%s"!', link)
        else:
            logging.error('Failed to load "%s"! (%d %s)', link, result.status_code, result.reason)
        return None

    return result.text


def download(link, dest, headers=None, random_ua=False):
    global HTTP_SESSION, DL_POOL, _DL_CANCEL

    if random_ua:
        if headers is None:
            headers = {}

        headers['User-Agent'] = get_user_agent(True)

    with DL_POOL:
        if _DL_CANCEL.is_set():
            return False

        logging.info('Downloading "%s"...', link)
        
        try:
            result = HTTP_SESSION.get(link, headers=headers, stream=True)
        except requests.exceptions.ConnectionError:
            logging.exception('Failed to load "%s"!', link)
            return False

        if result.status_code == 304:
            return 304
        elif result.status_code == 206:
            # sectorgame.com/fsfiles/ always returns code 206 which makes this necessary.
            logging.warning('"%s" returned "206 Partial Content", the downloaded file might be incomplete.', link)
        elif result.status_code != 200:
            logging.error('Failed to load "%s"! (%d %s)', link, result.status_code, result.reason)
            return False

        try:
            size = float(result.headers.get('content-length', 0))
        except:
            logging.exception('Failed to parse Content-Length header!')
            size = 1024 ** 4  # = 1 TB

        try:
            start = dest.tell()
            sc = SpeedCalc()
            for chunk in result.iter_content(512 * 1024):
                dest.write(chunk)

                if sc.push(dest.tell()) != -1:
                    if size > 0:
                        by_done = dest.tell() - start
                        speed = sc.get_speed()
                        p = by_done / size
                        text = ', ' + format_bytes(speed) + '/s, '
                        text += time.strftime('%M:%S', time.gmtime((size - by_done) / speed)) + ' left'
                    else:
                        p = 0
                        text = ''
                    progress.update(p, os.path.basename(link) + text)

                if _DL_CANCEL.is_set():
                    return False
        except:
            logging.exception('Download of "%s" was interrupted!', link)
            return False

    return True


def cancel_downloads():
    global _DL_CANCEL, DL_POOL

    _DL_CANCEL.set()
    # Wait for the downloads to actually cancel.
    while DL_POOL.get_consumed() > 0:
        time.sleep(0.2)

    _DL_CANCEL.clear()


# Tries all passed links until one succeeds.
def try_download(links, dest):
    for url in links:
        if download(url, dest):
            return True
    
    return False


# Actually transforms a given path into a platform-specific one.
def normpath(path):
    return os.path.normcase(path.replace('\\', '/'))


# Try to map a case insensitive path to an existing one.
def ipath(path):
    if os.path.exists(path):
        return path
    
    parent = os.path.dirname(path)
    if not os.path.exists(parent):
        parent = ipath(parent)

        if not os.path.exists(parent):
            # Well, nothing we can do here...
            return path

    siblings = os.listdir(parent)
    l_siblings = [s.lower() for s in siblings]
    
    oitem = item = os.path.basename(path)
    if item.lower() in l_siblings:
        item = siblings[l_siblings.index(item.lower())]
        path = os.path.join(parent, item)
    
    if item != oitem:
        logging.debug('Picking "%s" for "%s".', item, oitem)
    
    return path


# TODO: Shouldn't we also handle ./ and ../ here ?
def pjoin(*args):
    path = ''
    for arg in args:
        if arg.startswith('/'):
            path = arg
        elif path == '' or path.endswith('/'):
            path += arg
        else:
            path += '/' + arg
    
    return path


def url_join(a, b):
    if re.match(r'[a-zA-Z]+://.*', b):
        # A full URL
        return b

    if b == '':
        # Umm....
        return a

    if b[0] == '/':
        if len(b) > 1 and b[1] == '/':
            # The second part begins with // which means we have to grab a's protocol.
            proto = a[:a.find(':')]
            return proto + ':' + b

        # An absolute path
        info = re.match(r'([a-zA-Z]+://[^/]+).*', a)
        return info.group(1) + b

    return pjoin(a, b)


def gen_hash(path, algo='md5'):
    global HASH_CACHE
    
    path = os.path.abspath(path)
    info = os.stat(path)

    if algo == 'md5' and path in HASH_CACHE:
        chksum, mtime = HASH_CACHE[path]
        if mtime == info.st_mtime:
            # logging.debug('Found checksum for %s in cache.', path)
            return chksum
    
    logging.debug('Calculating checksum for %s...', path)

    h = hashlib.new(algo)
    with open(path, 'rb') as stream:
        while True:
            chunk = stream.read(16 * h.block_size)
            if not chunk:
                break

            h.update(chunk)

    chksum = h.hexdigest()
    if algo == 'md5':
        HASH_CACHE[path] = (chksum, info.st_mtime)
    
    return chksum


def test_7z():
    global SEVEN_PATH

    try:
        return call([SEVEN_PATH, '-h'], stdout=subprocess.DEVNULL) == 0
    except:
        logging.exception('Call to 7z failed!')

        if SEVEN_PATH == '7za':
            return False

        try:
            if call(['7za', '-h'], stdout=subprocess.DEVNULL) == 0:
                SEVEN_PATH = '7za'
                return True
            else:
                return False
        except:
            return False


def is_archive(path):
    path = path.lower()
    
    for ext in ARCHIVE_FORMATS:
        if path.endswith('.' + ext):
            return True
    
    return False


def extract_archive(archive, outpath, overwrite=False, files=None, _rec=False):
    global _HAS_TAR

    if archive.endswith(('.tar.gz', '.tar.xz', '.tar.bz2', '.tgz')):
        if _HAS_TAR is None:
            _HAS_TAR = call(['tar', '--version'], stdout=subprocess.DEVNULL) == 0

        if _HAS_TAR:
            cmd = ['tar', '-xf', archive, '-C', outpath]

            if archive.endswith('.gz'):
                cmd.append('-z')
            elif archive.endswith('.xz'):
                cmd.append('-J')
            elif archive.endswith('.bz2'):
                cmd.append('-j')

            if overwrite:
                cmd.append('--overwrite')

            if files:
                cmd.extend(files)

            if not os.path.isdir(outpath):
                os.makedirs(outpath)

            if QUIET:
                return call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
            else:
                return call(cmd) == 0

        if not _rec:
            # This is a file like whatever.tar.gz. We have to call 7z two times for this kind of file:
            # First to get whatever.tar and a second time to extract that tar archive.
            
            if not extract_archive(archive, os.path.dirname(archive), True, None, True):
                return False
            
            unc_archive = archive.split('.')
            # Remove the .gz or .bz2 or whatever ending...
            if unc_archive.pop() == 'tgz':
                unc_archive.append('tar')
            
            # ... and put it together again.
            unc_archive = '.'.join(unc_archive)
            res = extract_archive(unc_archive, outpath, overwrite, files, True)
            
            # Cleanup
            os.unlink(unc_archive)
            return res

    if archive.endswith('.dmg') and not _rec:
        # We have to call 7z twice for this file type. The first time, 7z only extracts the section contained in the file.
        # The second time, 7z extracts the actual contents from the HFS image.
        
        with tempfile.TemporaryDirectory() as tp:
            if not extract_archive(archive, tp, False, None, True):
                return False

            # Now look for the [0-9].hfs file. It could be that the number is always 3 but I want to be safe.
            image = glob.glob(os.path.join(tp, '*.hfs'))
            if len(image) < 1:
                return False

            # Now extract the image...
            return extract_archive(image[0], outpath, overwrite, files, True)
        
    cmd = [SEVEN_PATH, 'x', '-o' + outpath]
    if overwrite:
        cmd.append('-y')

    cmd.append(archive)

    if files is not None:
        cmd.extend(files)
    
    if QUIET:
        return call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
    else:
        return call(cmd) == 0


def convert_img(path, outfmt):
    global _HAS_CONVERT

    fd, dest = tempfile.mkstemp('.' + outfmt)
    os.close(fd)
    if Image is not None:
        img = Image.open(path)
        img.save(dest)
        
        return dest
    else:
        if _HAS_CONVERT is None:
            try:
                subprocess.check_call(['which', 'convert'], stdout=subprocess.DEVNULL)
                _HAS_CONVERT = True
            except subprocess.CalledProcessError:
                # Well, this failed, too. Is there any other way to convert an image?
                # For now I'll just abort.
                _HAS_CONVERT = False
                return None
        elif _HAS_CONVERT is False:
            return None
        
        subprocess.check_call(['convert', path, dest])
        return dest


def init_ui(ui, win):
    ui.setupUi(win)
    for attr in ui.__dict__:
        setattr(win, attr, getattr(ui, attr))

    return win


def is_number(s):
    try:
        int(s)
        return True
    except TypeError:
        return False


def merge_dicts(a, b):
    for k, v in b.items():
        if k in a and isinstance(v, dict) and isinstance(a[k], dict):
            merge_dicts(a[k], v)
        else:
            a[k] = v

    return a


def get_cpuinfo():
    from .launcher import get_cmd

    # Try the cpuid method first but do so in a seperate process in case it segfaults.
    try:
        info = json.loads(check_output(get_cmd(['--cpuinfo'])).strip())
    except subprocess.CalledProcessError:
        info = None
        logging.exception('The CPUID method failed!')
    except:
        info = None
        logging.exception('Failed to process my CPUID output.')

    if info is None:
        from .third_party import cpuinfo

        try:
            info = cpuinfo.get_cpu_info()
        except:
            logging.exception('Exception in the cpuinfo module!')
            info = None

    return info


def get_user_agent(random_ua=False):
    if random_ua:
        return random.choice(USER_AGENTS[1:])
    else:
        return USER_AGENTS[0]


def str_random(slen):
    s = ''
    for i in range(0, slen):
        s += random.choice('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789')

    return s


def human_list(items):
    if not isinstance(items, list):
        items = list(items)
    
    if len(items) == 0:
        return ''
    elif len(items) == 1:
        return items[0]
    else:
        return ', '.join(items[:-1]) + ' and ' + items[-1]


def connect(sig, cb, *args):
    cb = functools.partial(cb, *args)
    if not hasattr(sig, '_pyl'):
        sig._pyl = []

    sig._pyl.append(cb)
    sig.connect(cb)


class Spec(semantic_version.Spec):
    
    @classmethod
    def parse(self, specs_string):
        spec_texts = specs_string.split(',')
        res = []

        for spec_text in spec_texts:
            if '-' not in spec_text and '+' not in spec_text and spec_text != '*':
                spec_text = spec_text.split('.')
                while len(spec_text) < 3:
                    spec_text.append('0')

                spec_text = '.'.join(spec_text) + '-'

            res.append(semantic_version.SpecItem(spec_text))

        return tuple(res)

    @staticmethod
    def from_version(version, op='=='):
        return Spec('==' + str(version))


DL_POOL = ResizableSemaphore(10)
HTTP_SESSION.headers['User-Agent'] = get_user_agent()

if not center.DEBUG:
    logging.getLogger('requests.packages.urllib3.connectionpool').propagate = False

if sys.hexversion >= 0x020709F0:
    # In Python 2.7.9 the ssl module from Python 3.x was backported which means we can get rid of
    # the pyOpenSSL monkeypatch in urllib3.
    # On MacOS pyOpenSSL should be avoided at all cost since it's linked (by default) against
    # the system OpenSSL which is *very* old and doesn't support newer features like TLSv1.2.
    # NOTE: It might be better to check ssl.OPENSSL_VERSION
    # See also https://www.python.org/dev/peps/pep-0466/
    
    if 'requests.packages.urllib3.contrib.pyopenssl' in sys.modules:
        logging.info('Deactivating pyOpenSSL...')
        import requests.packages.urllib3.contrib.pyopenssl as pssl
        
        pssl.extract_from_urllib3()
        del pssl

if sys.platform.startswith('win'):
    _HAS_CONVERT = False
    _HAS_TAR = False



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/flags.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(455, 507)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.cmdLine = QtWidgets.QPlainTextEdit(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.cmdLine.sizePolicy().hasHeightForWidth())
        self.cmdLine.setSizePolicy(sizePolicy)
        self.cmdLine.setReadOnly(True)
        self.cmdLine.setObjectName("cmdLine")
        self.verticalLayout.addWidget(self.cmdLine)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.label_4 = QtWidgets.QLabel(Dialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.customFlags = QtWidgets.QLineEdit(Dialog)
        self.customFlags.setObjectName("customFlags")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.customFlags)
        self.easySetup = QtWidgets.QComboBox(Dialog)
        self.easySetup.setObjectName("easySetup")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.easySetup)
        self.listType = QtWidgets.QComboBox(Dialog)
        self.listType.setObjectName("listType")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.listType)
        self.verticalLayout.addLayout(self.formLayout)
        self.flagList = QtWidgets.QListWidget(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(4)
        sizePolicy.setHeightForWidth(self.flagList.sizePolicy().hasHeightForWidth())
        self.flagList.setSizePolicy(sizePolicy)
        self.flagList.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.flagList.setProperty("showDropIndicator", False)
        self.flagList.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.flagList.setObjectName("flagList")
        self.verticalLayout.addWidget(self.flagList)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.okButton = QtWidgets.QPushButton(Dialog)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        self.saveButton = QtWidgets.QPushButton(Dialog)
        self.saveButton.setObjectName("saveButton")
        self.horizontalLayout.addWidget(self.saveButton)
        self.defaultsButton = QtWidgets.QPushButton(Dialog)
        self.defaultsButton.setObjectName("defaultsButton")
        self.horizontalLayout.addWidget(self.defaultsButton)
        self.cancelButton = QtWidgets.QPushButton(Dialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Mod Flags"))
        self.label.setText(_translate("Dialog", "The complete commandline:"))
        self.label_2.setText(_translate("Dialog", "Easy Setup"))
        self.label_3.setText(_translate("Dialog", "Custom Flags"))
        self.label_4.setText(_translate("Dialog", "List Type"))
        self.okButton.setText(_translate("Dialog", "OK"))
        self.saveButton.setText(_translate("Dialog", "Save"))
        self.defaultsButton.setText(_translate("Dialog", "Use Defaults"))
        self.cancelButton.setText(_translate("Dialog", "Cancel"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings2.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(575, 400)
        Dialog.setModal(True)
        self.layout = QtWidgets.QHBoxLayout(Dialog)
        self.layout.setObjectName("layout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.saveButton = QtWidgets.QPushButton(Dialog)
        self.saveButton.setObjectName("saveButton")
        self.verticalLayout.addWidget(self.saveButton)
        self.treeWidget = QtWidgets.QTreeWidget(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.treeWidget.sizePolicy().hasHeightForWidth())
        self.treeWidget.setSizePolicy(sizePolicy)
        self.treeWidget.setMaximumSize(QtCore.QSize(149, 16777215))
        self.treeWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.treeWidget.setProperty("showDropIndicator", False)
        self.treeWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.treeWidget.setWordWrap(True)
        self.treeWidget.setHeaderHidden(True)
        self.treeWidget.setObjectName("treeWidget")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        self.treeWidget.topLevelItem(0).setText(0, "About Knossos")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        self.treeWidget.topLevelItem(1).setText(0, "Launcher settings")
        item_0.setFlags(QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(1).child(0).setText(0, "Retail install")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(1).child(1).setText(0, "Mod sources")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(1).child(2).setText(0, "Mod versions")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        self.treeWidget.topLevelItem(2).setText(0, "Game settings")
        item_0.setFlags(QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(2).child(0).setText(0, "Video")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(2).child(1).setText(0, "Audio")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(2).child(2).setText(0, "Input")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(2).child(3).setText(0, "Network")
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidget.topLevelItem(2).child(4).setText(0, "Default flags")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        self.treeWidget.topLevelItem(3).setText(0, "Help")
        self.verticalLayout.addWidget(self.treeWidget)
        self.layout.addLayout(self.verticalLayout)
        self.currentTab = QtWidgets.QWidget(Dialog)
        self.currentTab.setObjectName("currentTab")
        self.layout.addWidget(self.currentTab)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Settings"))
        self.saveButton.setText(_translate("Dialog", "Save"))
        self.treeWidget.headerItem().setText(0, _translate("Dialog", "1"))
        __sortingEnabled = self.treeWidget.isSortingEnabled()
        self.treeWidget.setSortingEnabled(False)
        self.treeWidget.setSortingEnabled(__sortingEnabled)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/log_viewer.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(648, 661)
        Dialog.setModal(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.pathLabel = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.pathLabel.setFont(font)
        self.pathLabel.setObjectName("pathLabel")
        self.verticalLayout.addWidget(self.pathLabel)
        self.content = QtWidgets.QTextBrowser(Dialog)
        self.content.setObjectName("content")
        self.verticalLayout.addWidget(self.content)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "fs2_open.log"))
        self.pathLabel.setText(_translate("Dialog", "fs2_open.log"))




## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import json
import re
import shutil
import hashlib
import semantic_version
import six
from . import center, util

# You have to fill this using https://github.com/workhorsy/py-cpuinfo .
CPU_INFO = None


class ModNotFound(Exception):
    mid = None
    spec = None

    def __init__(self, message, mid, spec=None):
        super(ModNotFound, self).__init__(message)

        self.mid = mid
        self.spec = spec


class PackageNotFound(ModNotFound):
    package = None

    def __init__(self, message, mid, package, spec=None):
        super(PackageNotFound, self).__init__(message, mid, spec)

        self.package = package


class Repo(object):
    base = None
    is_link = False
    mods = None
    includes = None
    pins = None

    def __init__(self, data=None):
        self.mods = {}
        self.pins = {}

        if data is not None:
            self.set(data)

    def empty(self):
        return len(self.mods) == 0

    def clear(self):
        self.mods = {}
        self.pins = {}

    def load_json(self, path):
        self.base = os.path.dirname(path)
        with open(path, 'r') as stream:
            self.set(json.load(stream))

    def save_json(self, path):
        with open(path, 'w') as stream:
            json.dump(self.get(), stream)

    def set(self, info):
        for m, v in info['pins'].items():
            self.pins[m] = semantic_version.Version(v)

        for mod in info['mods']:
            self.add_mod(Mod(mod, self))

    def get(self):
        mods = []
        for v in self.mods.values():
            for mod in v:
                mods.append(mod.get())

        pins = self.pins.copy()
        for m, v in pins.items():
            pins[m] = str(v)

        return {
            'pins': pins,
            'mods': mods
        }

    def fetch(self, link):
        self.base = os.path.dirname(link)
        self.is_link = True
        self.parse(util.get(link))

    def read(self, path):
        self.base = os.path.dirname(path)
        h = open(path, 'r')
        self.parse(h)
        h.close()

    def parse(self, obj):
        if not obj:
            return

        if isinstance(obj, six.string_types):
            data = json.loads(obj)
        else:
            data = json.load(obj)

        self.mods = {}
        self.includes = data.get('includes', [])

        for inc in self.includes:
            item = Repo()
            if self.is_link:
                item.fetch(util.url_join(self.base, inc))
            else:
                item.read(os.path.join(self.base, inc))

            self.merge(item)

        for mod in data.get('mods', []):
            self.add_mod(Mod(mod, self))

    def add_mod(self, mod):
        mid = mod.mid

        if len(mod.packages) == 0:
            logging.warning('Mod %s is empty, ignoring it!', mod)
            return

        if mid in self.mods:
            inserted = False

            for i, item in enumerate(self.mods[mid]):
                if item.version == mod.version:
                    if mod._repo is None:
                        mod_base = 'None'
                    else:
                        mod_base = mod._repo.base

                    logging.info('Mod "%s" (%s) from "%s" overwrites an existing mod version!', mid, mod.version, mod_base)

                    self.mods[mid][i] = mod
                    inserted = True
                    break

            if not inserted:
                self.mods[mid].append(mod)
                self.mods[mid].sort(key=lambda m: m.version, reverse=True)
        else:
            self.mods[mid] = [mod]

        mod._repo = self

    def merge(self, repo):
        for mvs in repo.mods.values():
            for mod in mvs:
                self.add_mod(mod)

    def pin(self, mod, version=None):
        if isinstance(mod, Package):
            mod = mod.get_mod()

        if isinstance(mod, Mod):
            version = mod.version
            mod = mod.mid

        if not isinstance(mod, str) or not isinstance(version, semantic_version.Version):
            raise ValueError('%s is not a string or %s is not a valid version!' % (mod, version))

        self.pins[mod] = version

    def unpin(self, mod):
        if isinstance(mod, Package):
            mod = mod.get_mod().mid
        elif isinstance(mod, Mod):
            mod = mod.mid

        if mod in self.pins:
            del self.pins[mod]

    def get_pin(self, mod):
        if isinstance(mod, Package):
            mod = mod.get_mod()

        if isinstance(mod, Mod):
            mod = mod.mid

        return self.pins.get(mod, None)

    def query(self, mid, spec=None, pname=None):
        if isinstance(mid, Package):
            mod = mid.get_mod()
            pname = mid.name
            mid = mod.mid
            spec = util.Spec.from_version(mod.version)

            del mod
        elif isinstance(mid, Mod):
            spec = util.Spec.from_version(mid.version)
            mid = mid.mid

        if mid not in self.mods:
            raise ModNotFound('Mod "%s" wasn\'t found!' % (mid), mid)

        if spec is None and mid in self.pins:
            spec = self.pins[mid]

        candidates = self.mods[mid]
        if spec is None:
            mod = candidates[0]
        else:
            version = spec.select([mod.version for mod in candidates])
            if not version:
                raise ModNotFound('Mod "%s" %s wasn\'t found!' % (mid, spec), mid, spec)

            for m in candidates:
                if m.version == version:
                    mod = m
                    break

        if pname is not None:
            for pkg in mod.packages:
                if pkg.name == pname:
                    return pkg

            raise ModNotFound('The package "%s" for mod "%s" wasn\'t found!' % (pname, mid), mid)
        else:
            return mod

    def query_all(self, mid, spec=None):
        if mid not in self.mods:
            raise ModNotFound('Mod "%s" wasn\'t found!' % (mid), mid)

        for mod in self.mods[mid]:
            if spec is None or spec.match(mod.version):
                yield mod

    def has(self, mid, spec=None, pname=None):
        try:
            self.query(mid, spec, pname)
            return True
        except ModNotFound:
            return False

    def get_tree(self):
        submod_ids = []
        for mid, mod in self.mods.items():
            submod_ids.extend(mod[0].submods)

        roots = set(self.mods.keys()) - set(submod_ids)
        return [self.mods[mid][0] for mid in roots]

    def get_list(self):
        for mod in self.mods.values():
            yield mod[0]

    # TODO: Is this overcomplicated?
    def process_pkg_selection(self, pkgs):
        dep_dict = {}
        ndeps = pkgs

        # Resolve the pkgs' dependencies
        while len(ndeps) > 0:
            _nd = ndeps
            ndeps = []

            for pkg in _nd:
                for dep, version in pkg.resolve_deps():
                    mid = dep.get_mod().mid

                    if mid not in dep_dict:
                        dep_dict[mid] = {}

                    if dep.name not in dep_dict[mid]:
                        dep_dict[mid][dep.name] = {}

                    if version not in dep_dict[mid][dep.name]:
                        dep_dict[mid][dep.name][version] = dep
                        ndeps.append(dep)

        # Check for conflicts (and try to resolve them if possible).
        dep_list = set()
        for mid, deps in dep_dict.items():
            for name, variants in deps.items():
                if len(variants) == 1:
                    dep_list.add(next(iter(variants.values())))
                else:
                    specs = variants.keys()
                    remains = []

                    for v in variants.values():
                        ok = True
                        for spec in specs:
                            if not spec.match(v.get_mod().version):
                                ok = False
                                break

                        if ok:
                            remains.append(v)

                    if len(remains) == 0:
                        v = (list(variants.values())[0].name, ','.join([str(s) for s in specs]))
                        raise PackageNotFound('No version of package "%s" found for these constraints: %s' % v, mid, list(variants.values())[0].name)
                    else:
                        # Pick the latest
                        remains.sort(key=lambda v: v.get_mod().version)
                        dep_list.add(remains[-1])

        dep_list |= set(pkgs)
        return dep_list

    def save_logos(self, path):
        for mid, mvs in self.mods.items():
            for mod in mvs:
                mod.save_logo(path)


class Mod(object):
    _repo = None
    mid = ''
    title = ''
    mtype = 'mod'
    version = None
    folder = None
    cmdline = ''
    logo = None
    logo_path = None
    description = ''
    notes = ''
    submods = None
    actions = None
    packages = None

    __fields__ = ('mid', 'title', 'type', 'version', 'folder', 'cmdline', 'logo', 'description', 'notes', 'submods', 'actions', 'packages')

    def __init__(self, values=None, repo=None):
        self.actions = []
        self.packages = []

        if repo is not None:
            self._repo = repo

        if values is not None:
            self.set(values)

    def __repr__(self):
        return '<Mod "%s" %s (%s)>' % (self.title, self.version, self.mid)

    def set(self, values):
        self.mid = values['id']
        self.title = values['title']
        self.mtype = values.get('type', 'mod')  # Backwards compatibility
        self.version = semantic_version.Version(values['version'], partial=True)
        self.folder = values.get('folder', self.mid).strip('/')  # make sure we have a relative path
        self.cmdline = values.get('cmdline', '')
        self.logo = values.get('logo', None)
        self.description = values.get('description', '')
        self.notes = values.get('notes', '')
        self.submods = values.get('submods', [])
        self.actions = values.get('actions', [])

        self.packages = []
        if self.mtype != 'mod' or center.has_retail:
            for pkg in values.get('packages', []):
                pkg = Package(pkg, self)
                if pkg.check_env():
                    self.packages.append(pkg)

        if self._repo is not None and self._repo.base is not None and self.logo is not None:
            if '://' in self._repo.base:
                self.logo = util.url_join(self._repo.base, self.logo)
            else:
                self.logo = os.path.abspath(os.path.join(self._repo.base, self.logo))

        # Enforce relative paths
        for act in self.actions:
            if 'paths' in act:
                act['paths'] = [p.lstrip('/') for p in act['paths']]

            if 'dest' in act:
                act['dest'] = act['dest'].lstrip('/')

    def get(self):
        return {
            'id': self.mid,
            'title': self.title,
            'type': self.mtype,
            'version': str(self.version),
            'folder': self.folder,
            'cmdline': self.cmdline,
            'logo': self.logo,
            'description': self.description,
            'notes': self.notes,
            'submods': self.submods,
            'actions': self.actions,
            'packages': [pkg.get() for pkg in self.packages]
        }

    def copy(self):
        return Mod(self.get(), self._repo)

    def get_submods(self):
        return [self._repo.query(mid) for mid in self.submods]

    def get_files(self):
        files = []
        for pkg in self.packages:
            for item in pkg.filelist:
                item = item.copy()
                item['package'] = pkg.name
                files.append(item)

        return files

    def resolve_deps(self, only_required=True):
        if only_required:
            pkgs = [pkg for pkg in self.packages if pkg.status == 'required']
        else:
            pkgs = [pkg for pkg in self.packages if pkg.status in ('required', 'recommended')]

        return self._repo.process_pkg_selection(pkgs)

    def save_logo(self, dest):
        if self.logo is None:
            return

        suffix = '.' + self.logo.split('.')[-1]
        path = os.path.join(dest, 'logo_' + hashlib.md5(self.logo.encode('utf8')).hexdigest() + suffix)

        if not os.path.isfile(path):
            if '://' in self.logo:
                # That's a URL
                with open(path, 'wb') as fobj:
                    util.download(self.logo, fobj)
            else:
                shutil.copyfile(self.logo, path)

        self.logo = os.path.abspath(path)


class Package(object):
    _mod = None
    name = ''
    notes = ''
    status = 'recommended'
    dependencies = None
    environment = None
    files = None
    filelist = None
    executables = None

    def __init__(self, values=None, mod=None):
        self._mod = mod
        self.dependencies = []
        self.environment = []
        self.files = {}

        if values is not None:
            self.set(values)

    def __repr__(self):
        return '<Package "%s" of %s>' % (self.name, self._mod)

    def set(self, values):
        self.name = values['name']
        self.notes = values.get('notes', '')
        self.status = values.get('status', 'recommended').lower()
        self.dependencies = values.get('dependencies', [])
        self.environment = values.get('environment', [])
        self.files = {}
        self.filelist = values.get('filelist', [])
        self.executables = values.get('executables', [])

        _files = values.get('files', [])

        if isinstance(_files, dict):
            self.files = _files
            for name, item in _files.items():
                item['filename'] = name
                item['dest'] = item.get('dest', '').strip('/')  # make sure this is a relative path

        elif isinstance(_files, list):
            for item in _files:
                item['dest'] = item.get('dest', '').strip('/')  # make sure this is a relative path
                self.files[item['filename']] = item
        else:
            logging.warning('"%s"\'s file list has an unknown type.', self.name)

        has_mod_dep = False
        mid = self._mod.mid

        if mid == '':
            raise Exception('Package "%s" initialized with Mod %s which has no ID!' % (self.name, self._mod.title))

        for info in self.dependencies:
            if info['id'] == mid:
                has_mod_dep = True
                break

        if not has_mod_dep:
            self.dependencies.append({
                'id': self.get_mod().mid,
                'version': '==' + str(self.get_mod().version),
                'packages': []
            })

    def get(self):
        return {
            'name': self.name,
            'notes': self.notes,
            'status': self.status,
            'dependencies': self.dependencies,
            'environment': self.environment,
            'files': list(self.files.values()),
            'filelist': self.filelist,
            'executables': self.executables
        }

    def get_mod(self):
        return self._mod

    def get_files(self):
        files = {}
        for name, item in self.files.items():
            if item['is_archive']:
                for path, csum in item['contents'].items():
                    files[os.path.join(item['dest'], path)] = (csum, name)
            else:
                files[os.path.join(item['dest'], name)] = (item['md5sum'], name)

        return files

    def resolve_deps(self):
        result = []
        for dep in self.dependencies:
            version = dep['version']
            if re.match('\d.*', version):
                # Make a spec out of this version
                version = '==' + version

            version = util.Spec(version)
            mod = self._mod._repo.query(dep['id'], version)
            pkgs = dep.get('packages', [])
            found_pkgs = []

            for pkg in mod.packages:
                if pkg.status == 'required' or pkg.name in pkgs:
                    result.append((pkg, version))
                    found_pkgs.append(pkg.name)

            missing_pkgs = set(pkgs) - set(found_pkgs)
            if len(missing_pkgs) > 0:
                raise PackageNotFound('Package %s of mod %s (%s) couldn\'t be found!' % (next(iter(missing_pkgs)), mod.mid, version), mod.mid, missing_pkgs)

        return result

    def check_env(self):
        bvars = None

        for check in self.environment:
            value = check['value'].lower()
            c_type = check['type'].lower()

            if c_type == 'os':
                if value == 'windows':
                    if sys.platform not in ('win32', 'cygwin'):
                        return False
                elif value == 'linux':
                    if not sys.platform.startswith('linux'):
                        return False
                elif value == 'macos':
                    if sys.platform != 'darwin':
                        return False
                else:
                    return False

            elif c_type == 'cpu_feature':
                # return CPU_INFO is None or value in CPU_INFO['flags']
                if CPU_INFO is None:
                    # We don't have any information on the current CPU so we just ignore this check.
                    return True

                if value in CPU_INFO['flags']:
                    return True

                if value.upper() == CPU_INFO['arch']:
                    return True

                return False
            elif c_type == 'bool':
                if bvars is None:
                    bvars = {}
                    bvars[CPU_INFO['arch']] = True  # this is either X86_32 or X86_64

                    if sys.platform in ('win32', 'cygwin'):
                        bvars['windows'] = True
                    elif sys.platform.startswith('linux'):
                        bvars['linux'] = True
                    elif sys.platform == 'darwin':
                        bvars['macosx'] = True
                    else:
                        logging.error('You are using an unrecognized OS! (%s)' % sys.platform)

                    for flag in CPU_INFO['flags']:
                        bvars[flag] = True

                return self._solve_bool(value, bvars)

        return True

    @classmethod
    def _solve_bool(cls, expr, vars):
        if expr[0] == 'var':
            return vars.get(expr[1])
        elif expr[0] == 'not':
            return not cls._solve_bool(expr[1], vars)
        elif expr[0] == 'and':
            return cls._solve_bool(expr[1], vars) and cls._solve_bool(expr[2], vars)
        elif expr[0] == 'or':
            return cls._solve_bool(expr[1], vars) or cls._solve_bool(expr[2], vars)
        else:
            raise Exception('Unknown operation %s! (%s)' % (expr[0], expr))


# Keeps track of installed mods
class InstalledRepo(Repo):
    base = '[INSTALLED]'

    def clear(self):
        self.mods = {}

    def save_pins(self, path):
        with open(path, 'w') as stream:
            json.dump(self.pins, stream)

    def load_pins(self, path):
        with open(path, 'r') as stream:
            self.pins = json.load(stream)

    def set(self, mods):
        for m, v in mods['pins'].items():
            self.pins[m] = semantic_version.Version(v)

        for mod in mods['mods']:
            self.add_mod(InstalledMod(mod))

    def add_pkg(self, pkg):
        mod = pkg.get_mod()
        try:
            my_mod = self.query(mod)
        except ModNotFound:
            my_mod = InstalledMod.convert(mod)
            my_pkg = my_mod.add_pkg(pkg)
            self.add_mod(my_mod)
            return my_pkg
        else:
            return my_mod.add_pkg(pkg)

    def del_pkg(self, pkg):
        mod = pkg.get_mod()

        try:
            my_mod = self.query(mod)
        except ModNotFound:
            logging.error('Tried to delete non-existing package!')
            return

        my_mod.del_pkg(pkg)

    def del_mod(self, mod):
        if mod.mid in self.mods:
            vs = self.mods[mod.mid]
            rem = None
            for m in vs:
                if m.version == mod.version:
                    rem = m
                    break

            if rem is None:
                logging.error('Tried to delete missing mod version!')
            else:
                vs.remove(rem)
                if len(vs) == 0:
                    del self.mods[mod.mid]

    def is_installed(self, mid, spec=None, pname=None):
        try:
            self.query(mid, spec, pname)
            return True
        except ModNotFound:
            return False

    def get_updates(self):
        remote_mods = center.mods
        updates = {}

        for mid, mods in self.mods.items():
            try:
                rem_mod = remote_mods.query(mid)
            except ModNotFound:
                continue

            if rem_mod.version > mods[0].version:
                # Let's see if the files changed.
                my_files = {}
                rem_files = {}
                my_pkgs = [pkg.name for pkg in mods[0].packages]

                for item in mods[0].get_files():
                    my_files[item['filename']] = item['md5sum']

                for item in rem_mod.get_files():
                    if item['package'] in my_pkgs:
                        rem_files[item['filename']] = item['md5sum']

                if rem_files == my_files:
                    logging.warning('Detected an empty update for mod "%s"! (%s -> %s)', mods[0].title, str(mods[0].version), str(rem_mod.version))
                    # TODO: Resolve this situation! (Update the local metadata?)
                else:
                    if mid not in updates:
                        updates[mid] = {}

                    updates[mid][mods[0].version] = rem_mod.version

        return updates


class InstalledMod(Mod):
    check_notes = ''

    @staticmethod
    def load(path):
        if path.endswith('.json'):
            with open(path, 'r') as stream:
                data = json.load(stream)

            mod = InstalledMod(data)
        elif path.endswith('.ini'):
            mod = IniMod()
            mod.load(path)
        else:
            return None

        mod.folder = os.path.basename(os.path.dirname(path))
        if mod.logo is not None and '://' not in mod.logo:
            mod.logo_path = os.path.join(os.path.dirname(path), mod.logo)

        return mod

    @staticmethod
    def convert(mod):
        data = mod.get()
        data['packages'] = []

        return InstalledMod(data)

    def __init__(self, values=None):
        super(InstalledMod, self).__init__(values)

    def set(self, values):
        pkgs = values.get('packages', [])
        values = values.copy()
        values['packages'] = []

        super(InstalledMod, self).set(values)

        self.check_notes = values.get('check_notes', '')
        for pkg in pkgs:
            self.packages.append(InstalledPackage(pkg, self))

    def get(self):
        return {
            'id': self.mid,
            'title': self.title,
            'version': str(self.version),
            'description': self.description,
            'logo': self.logo,
            'cmdline': self.cmdline,
            'packages': [pkg.get() for pkg in self.packages]
        }

    def add_pkg(self, pkg):
        pkg = InstalledPackage.convert(pkg, self)
        found = False

        for i, p in enumerate(self.packages):
            if p.name == pkg.name:
                self.packages[i] = pkg
                found = True
                break

        if not found:
            self.packages.append(pkg)

        return pkg

    def del_pkg(self, pkg):
        for i, p in enumerate(self.packages):
            if p.name == pkg.name:
                del self.packages[i]
                break

    def save(self):
        modpath = os.path.join(center.settings['fs2_path'], self.folder)
        path = os.path.join(modpath, 'mod.json')
        info = self.get()

        if self.logo is not None and not self.logo.startswith('knossos.'):
            logo = os.path.join(modpath, 'knossos.' + self.logo.split('.')[-1])

            if os.path.abspath(logo) != os.path.abspath(self.logo):
                # Copy the logo right next to the json file.
                shutil.copy(self.logo, logo)

            info['logo'] = os.path.basename(logo)

        with open(path, 'w') as stream:
            json.dump(info, stream)

    def get_mod_flag(self):
        mods = [self.folder]

        try:
            for dep in self.resolve_deps():
                folder = dep.get_mod().folder
                if folder not in mods:
                    mods.append(folder)
        except ModNotFound:
            logging.exception('A dependency for an installed mod is missing!')
            raise

        m = []
        for item in mods:
            if item.strip() != '':
                m.append(os.path.basename(util.ipath(os.path.join(center.settings['fs2_path'], item))))

        return m


class IniMod(InstalledMod):
    _pr_list = None
    _sc_list = None

    def __init__(self, values=None):
        super(IniMod, self).__init__(values=None)

        self._pr_list = []
        self._sc_list = []

        self.version = semantic_version.Version('1.0.0+ini')

    def load(self, path):
        with open(path, 'r') as stream:
            for line in stream:
                if '=' not in line:
                    continue

                line = line.split('=')
                name = line[0].strip()
                value = line[1].strip(' \r\n\t;')

                if name == 'modname':
                    self.title = value + ' (ini)'
                elif name == 'infotext':
                    self.description = value
                elif name.startswith('image'):
                    self.logo = value
                elif name == 'primarylist':
                    if value != '':
                        self._pr_list = value.split(',')
                elif name in ('secondarylist', 'secondrylist'):
                    if value != '':
                        self._sc_list = value.split(',')

        folder = os.path.basename(os.path.dirname(path))
        if self.title == '':
            self.title = folder + ' (ini)'

        self.mid = '##INI_COMPAT#' + folder

        pkg = InstalledPackage({
            'name': 'Content',
            'status': 'required'
        }, self)
        self.add_pkg(pkg)

    def get_mod_flag(self):
        mods = self._pr_list[:]
        mods.append(self.folder)
        mods.extend(self._sc_list)

        return mods


class InstalledPackage(Package):
    check_notes = ''
    files_ok = -1
    files_checked = -1

    @staticmethod
    def convert(pkg, mod):
        return InstalledPackage(pkg.get(), mod)

    def set(self, values):
        super(InstalledPackage, self).set(values.copy())

        self.check_notes = values.get('check_notes', '')

    def get(self):
        data = super(InstalledPackage, self).get()
        data['check_notes'] = self.check_notes
        return data



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(425, 626)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setWordWrap(False)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.build = QtWidgets.QComboBox(Dialog)
        self.build.setObjectName("build")
        self.horizontalLayout_3.addWidget(self.build)
        self.browseButton = QtWidgets.QPushButton(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.browseButton.sizePolicy().hasHeightForWidth())
        self.browseButton.setSizePolicy(sizePolicy)
        self.browseButton.setObjectName("browseButton")
        self.horizontalLayout_3.addWidget(self.browseButton)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.line = QtWidgets.QFrame(Dialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.label_2 = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.formLayout_4 = QtWidgets.QFormLayout()
        self.formLayout_4.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_4.setObjectName("formLayout_4")
        self.resolutionLabel = QtWidgets.QLabel(Dialog)
        self.resolutionLabel.setObjectName("resolutionLabel")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.resolutionLabel)
        self.vid_res = QtWidgets.QComboBox(Dialog)
        self.vid_res.setObjectName("vid_res")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.vid_res)
        self.depthLabel = QtWidgets.QLabel(Dialog)
        self.depthLabel.setObjectName("depthLabel")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.depthLabel)
        self.vid_depth = QtWidgets.QComboBox(Dialog)
        self.vid_depth.setObjectName("vid_depth")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.vid_depth)
        self.horizontalLayout_2.addLayout(self.formLayout_4)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.textureFilterLabel = QtWidgets.QLabel(Dialog)
        self.textureFilterLabel.setObjectName("textureFilterLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.textureFilterLabel)
        self.vid_texfilter = QtWidgets.QComboBox(Dialog)
        self.vid_texfilter.setObjectName("vid_texfilter")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.vid_texfilter)
        self.antialiasingLabel = QtWidgets.QLabel(Dialog)
        self.antialiasingLabel.setObjectName("antialiasingLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.antialiasingLabel)
        self.vid_aa = QtWidgets.QComboBox(Dialog)
        self.vid_aa.setObjectName("vid_aa")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.vid_aa)
        self.anisotropicFilterLabel = QtWidgets.QLabel(Dialog)
        self.anisotropicFilterLabel.setObjectName("anisotropicFilterLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.anisotropicFilterLabel)
        self.vid_af = QtWidgets.QComboBox(Dialog)
        self.vid_af.setObjectName("vid_af")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.vid_af)
        self.horizontalLayout_2.addLayout(self.formLayout)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.line_3 = QtWidgets.QFrame(Dialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.label_3 = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        self.formLayout_3 = QtWidgets.QFormLayout()
        self.formLayout_3.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_3.setObjectName("formLayout_3")
        self.playbackDeviceLabel = QtWidgets.QLabel(Dialog)
        self.playbackDeviceLabel.setObjectName("playbackDeviceLabel")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.playbackDeviceLabel)
        self.snd_playback = QtWidgets.QComboBox(Dialog)
        self.snd_playback.setObjectName("snd_playback")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.snd_playback)
        self.captureDeviceLabel = QtWidgets.QLabel(Dialog)
        self.captureDeviceLabel.setObjectName("captureDeviceLabel")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.captureDeviceLabel)
        self.snd_capture = QtWidgets.QComboBox(Dialog)
        self.snd_capture.setObjectName("snd_capture")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.snd_capture)
        self.verticalLayout.addLayout(self.formLayout_3)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.snd_efx = QtWidgets.QCheckBox(Dialog)
        self.snd_efx.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.snd_efx.setObjectName("snd_efx")
        self.horizontalLayout_5.addWidget(self.snd_efx)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.sampleRateLabel = QtWidgets.QLabel(Dialog)
        self.sampleRateLabel.setObjectName("sampleRateLabel")
        self.horizontalLayout_5.addWidget(self.sampleRateLabel)
        self.snd_samplerate = QtWidgets.QSpinBox(Dialog)
        self.snd_samplerate.setObjectName("snd_samplerate")
        self.horizontalLayout_5.addWidget(self.snd_samplerate)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.line_4 = QtWidgets.QFrame(Dialog)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.verticalLayout.addWidget(self.line_4)
        self.keyboardLabel = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.keyboardLabel.setFont(font)
        self.keyboardLabel.setObjectName("keyboardLabel")
        self.verticalLayout.addWidget(self.keyboardLabel)
        self.keyboardForm = QtWidgets.QFormLayout()
        self.keyboardForm.setObjectName("keyboardForm")
        self.label_8 = QtWidgets.QLabel(Dialog)
        self.label_8.setObjectName("label_8")
        self.keyboardForm.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.setxkbmapCheck = QtWidgets.QCheckBox(Dialog)
        self.setxkbmapCheck.setObjectName("setxkbmapCheck")
        self.keyboardForm.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.setxkbmapCheck)
        self.keyLayoutSelect = QtWidgets.QComboBox(Dialog)
        self.keyLayoutSelect.setObjectName("keyLayoutSelect")
        self.keyboardForm.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.keyLayoutSelect)
        self.verticalLayout.addLayout(self.keyboardForm)
        self.line_7 = QtWidgets.QFrame(Dialog)
        self.line_7.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_7.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_7.setObjectName("line_7")
        self.verticalLayout.addWidget(self.line_7)
        self.label_4 = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.ctrl_joystick = QtWidgets.QComboBox(Dialog)
        self.ctrl_joystick.setObjectName("ctrl_joystick")
        self.verticalLayout.addWidget(self.ctrl_joystick)
        self.line_6 = QtWidgets.QFrame(Dialog)
        self.line_6.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.verticalLayout.addWidget(self.line_6)
        self.label_6 = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.verticalLayout.addWidget(self.label_6)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_2.setObjectName("formLayout_2")
        self.connectionTypeLabel = QtWidgets.QLabel(Dialog)
        self.connectionTypeLabel.setObjectName("connectionTypeLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.connectionTypeLabel)
        self.net_type = QtWidgets.QComboBox(Dialog)
        self.net_type.setObjectName("net_type")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.net_type)
        self.connectionSpeedLabel = QtWidgets.QLabel(Dialog)
        self.connectionSpeedLabel.setObjectName("connectionSpeedLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.connectionSpeedLabel)
        self.net_speed = QtWidgets.QComboBox(Dialog)
        self.net_speed.setObjectName("net_speed")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.net_speed)
        self.horizontalLayout_4.addLayout(self.formLayout_2)
        self.formLayout_6 = QtWidgets.QFormLayout()
        self.formLayout_6.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_6.setObjectName("formLayout_6")
        self.forceLocalPortLabel = QtWidgets.QLabel(Dialog)
        self.forceLocalPortLabel.setObjectName("forceLocalPortLabel")
        self.formLayout_6.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.forceLocalPortLabel)
        self.net_port = QtWidgets.QLineEdit(Dialog)
        self.net_port.setObjectName("net_port")
        self.formLayout_6.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.net_port)
        self.forceIPAddressLabel = QtWidgets.QLabel(Dialog)
        self.forceIPAddressLabel.setObjectName("forceIPAddressLabel")
        self.formLayout_6.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.forceIPAddressLabel)
        self.net_ip = QtWidgets.QLineEdit(Dialog)
        self.net_ip.setObjectName("net_ip")
        self.formLayout_6.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.net_ip)
        self.horizontalLayout_4.addLayout(self.formLayout_6)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.line_5 = QtWidgets.QFrame(Dialog)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.verticalLayout.addWidget(self.line_5)
        self.label_5 = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.verticalLayout.addWidget(self.label_5)
        self.cmdButton = QtWidgets.QPushButton(Dialog)
        self.cmdButton.setObjectName("cmdButton")
        self.verticalLayout.addWidget(self.cmdButton)
        self.line_2 = QtWidgets.QFrame(Dialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.runButton = QtWidgets.QPushButton(Dialog)
        self.runButton.setDefault(True)
        self.runButton.setObjectName("runButton")
        self.horizontalLayout.addWidget(self.runButton)
        self.saveButton = QtWidgets.QPushButton(Dialog)
        self.saveButton.setObjectName("saveButton")
        self.horizontalLayout.addWidget(self.saveButton)
        self.cancelButton = QtWidgets.QPushButton(Dialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Settings"))
        self.label.setText(_translate("Dialog", "FreeSpace 2 Open build"))
        self.browseButton.setText(_translate("Dialog", "Browse"))
        self.label_2.setText(_translate("Dialog", "Video"))
        self.resolutionLabel.setText(_translate("Dialog", "Resolution :"))
        self.depthLabel.setText(_translate("Dialog", "Color depth :"))
        self.textureFilterLabel.setText(_translate("Dialog", "Texture filter :"))
        self.antialiasingLabel.setText(_translate("Dialog", "Antialiasing :"))
        self.anisotropicFilterLabel.setText(_translate("Dialog", "Anisotropic filtering :"))
        self.label_3.setText(_translate("Dialog", "Audio"))
        self.playbackDeviceLabel.setText(_translate("Dialog", "Playback device :"))
        self.captureDeviceLabel.setText(_translate("Dialog", "Capture device :"))
        self.snd_efx.setText(_translate("Dialog", "Enable EFX"))
        self.sampleRateLabel.setText(_translate("Dialog", "Sample rate :"))
        self.keyboardLabel.setText(_translate("Dialog", "Keyboard"))
        self.label_8.setText(_translate("Dialog", "Layout: "))
        self.setxkbmapCheck.setText(_translate("Dialog", "Use \"setxbmap\""))
        self.label_4.setText(_translate("Dialog", "Controller"))
        self.label_6.setText(_translate("Dialog", "Network"))
        self.connectionTypeLabel.setText(_translate("Dialog", "Connection type :"))
        self.connectionSpeedLabel.setText(_translate("Dialog", "Connection speed :"))
        self.forceLocalPortLabel.setText(_translate("Dialog", "Force local port :"))
        self.forceIPAddressLabel.setText(_translate("Dialog", "Force IP address :"))
        self.label_5.setText(_translate("Dialog", "Advanced settings"))
        self.cmdButton.setText(_translate("Dialog", "Command line flags for Mod name"))
        self.runButton.setText(_translate("Dialog", "Launch"))
        self.saveButton.setText(_translate("Dialog", "Save"))
        self.cancelButton.setText(_translate("Dialog", "Close"))

